# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS Custom step"""

import seamm
import Pmw
import tkinter as tk


class TkCustom(seamm.TkNode):
    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=None, y=None, w=200, h=50
    ):
        """Initialize a node

        Keyword arguments:
        """

        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(self):
        """Create the dialog!"""
        frame = super().create_dialog("Edit LAMMPS Custom Step")

        fixedFont = Pmw.logicalfont("Fixed")
        text = self._widget["text"] = Pmw.ScrolledText(
            frame,
            labelpos="n",
            label_text="Custom script for LAMMPS",
            text_wrap="none",
            text_font=fixedFont,
            text_padx=4,
            text_pady=4,
        )
        text.insert("1.0", self.node.parameters["script"].value)
        text.pack(expand=tk.YES, fill=tk.BOTH)

    def handle_dialog(self, result):
        """Do the right thing when the dialog is closed."""
        if result is None or result == "Cancel":
            self.dialog.deactivate(result)
            self["text"].delete(1.0, "end")
            self["text"].insert(1.0, self.node.parameters["script"].value)
        elif result == "Help":
            self.help()
        elif result == "OK":
            self.dialog.deactivate(result)
            # Capture the parameters from the widgets
            self.node.parameters["script"].value = (
                self["text"].get(1.0, tk.END).rstrip()
            )
        else:
            self.dialog.deactivate(result)
            raise RuntimeError("Don't recognize dialog result '{}'".format(result))



# -*- coding: utf-8 -*-

"""NPT (canonical) dynamics in LAMMPS"""

import json

import lammps_step
import logging
import seamm
from seamm_util import units_class, Q_
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import random

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("lammps")


class NPT(lammps_step.NVT):
    methods = {
        "Nose-Hoover": {
            "documentation": "https://lammps.sandia.gov/doc/fix_nh.html#fix-npt-command",  # noqa: E501
            "references": ["Shinoda", "Tuckerman"],
        },
        "Berendsen": {
            "documentation": "https://lammps.sandia.gov/doc/fix_press_berendsen.html",  # noqa: E501
            "references": ["Berendsen"],
        },
    }

    def __init__(self, flowchart=None, title="NPT dynamics", extension=None):
        """Initialize the node"""

        logger.debug("Creating NPT {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        logger.debug("NPT after super init, {}".format(self))

        self.description = "NPT dynamics step in LAMMPS"

        logger.debug("NPT.init() creating NPT_Parameters object")

        self._calculation = "npt"
        self._model = None
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.NPT_Parameters()

        logger.debug("NPT.init() completed")

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.
        """

        if not P:
            P = self.parameters.values_to_dict()

        # What will we do?

        if P["T0"] == P["T1"]:
            text = "{time} of canonical (NPT) dynamics at {T0} "
        else:
            text = (
                "{time} of canonical (NPT) dynamics starting "
                " at {T0}, going to {T1}, "
            )
        if P["thermostat"] == "Nose-Hoover":
            text += "using a Nose-Hoover thermostat."
            if P["Tchain"] != "3":
                if P["Tloop"] != "1":
                    text += (
                        " The thermostat will use a chain of {Tchain} "
                        "thermostats with {Tloop} subcycles and a "
                    )
                else:
                    text += (
                        " The thermostat will use a chain of {Tchain} "
                        "thermostats and a "
                    )
            elif P["Tloop"] != "1":
                text += " The thermostat will use {Tloop} subcycles and a "
            else:
                text += " The thermostat will use a "
            text += "drag factor of {drag}."
        elif P["thermostat"] == "Berendsen":
            text += "using a Berendsen thermostat with a damping time " "of {Tdamp}"
        elif "csvr" in P["thermostat"]:
            text += (
                "using a canonical sampling thermostat using velocity "
                "rescaling (CSVR) with a damping time of {Tdamp} and "
                "a {random_seed}."
            )
        elif "csld" in P["thermostat"]:
            text += (
                "using a canonical sampling thermostat using Langevin "
                "dynamics (CSLD) with a damping time of {Tdamp} and "
                "a {random_seed}."
            )
        elif P["thermostat"] == "velocity rescaling":
            text += (
                "using velocity rescaling every {frequency} with a "
                "temperature window of {window}."
            )
            if P["fraction"] != 1.0:
                text += (
                    " The velocities will only be scaled a fraction "
                    "({fraction}) of the amount needed to fully correct "
                    "the temperature."
                )
        elif P["thermostat"] == "Langevin":
            text += (
                "using a Langevin thermostat with a damping time "
                "of {Tdamp} and a {random_seed}"
            )
        else:
            text += "using the thermostat given by {thermostat}"

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def describe(self, indent="", json_dict=None):
        """Write out information about what this node will do
        If json_dict is passed in, add information to that dictionary
        so that it can be written out by the controller as appropriate.
        """

        # Can't call super() because it will print too much
        self.visited = True
        job.job("\n" + self.indent + self.header)
        next_node = self.next()

        # Local copies of variables in a dictionary

        P = self.parameters.values_to_dict()
        text = self.description_text(P)
        job.job(__(text, indent=self.indent + "    ", **P))

        return next_node

    def get_input(self, extras=None):
        """Get the input for an NPT dynamics run in LAMMPS"""

        keep_orthorhombic = True

        self.description = []

        _, configuration = self.get_system_configuration()
        if configuration.periodicity == 0:
            raise RuntimeError("Cannot run NPT dynamics on non-periodic system!")

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Fix variables with special cases
        timestep, P["timestep"] = self.timestep(P["timestep"])

        if P["seed"] == "random":
            # Apparently the seed must be no larger than 900,000,000
            P["seed"] = int(random.random() * 899999999)

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description.append(__(self.description_text(PP), **PP, indent=3 * " "))

        time = lammps_step.to_lammps_units(P["time"], quantity="time")
        nsteps = round(time / timestep)

        T0 = lammps_step.to_lammps_units(P["T0"], quantity="temperature")
        T1 = lammps_step.to_lammps_units(P["T1"], quantity="temperature")
        Tdamp = lammps_step.to_lammps_units(P["Tdamp"], quantity="time")

        barostat = P["barostat"]
        if barostat == "Berendsen":
            modulus = lammps_step.to_lammps_units(P["modulus"], quantity="pressure")

        # Work out the pressure/stress part of the command
        ptext = self.get_pressure_text(P, keep_orthorhombic)

        thermo_properties = (
            "time temp press etotal ke pe ebond "
            "eangle edihed eimp evdwl etail ecoul elong"
        )
        properties = (
            "v_time v_temp v_press v_vol v_density v_cella v_cellb "
            "v_cellc v_etotal v_ke v_pe v_epair"
        )
        title2 = "tstep t T P V density a b c Etot Eke Epe Epair"

        # and build the LAMMPS script
        lines = []
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append("reset_timestep      0")
        lines.append("timestep            {}".format(timestep))
        lines.append("thermo_style        custom {}".format(thermo_properties))
        lines.append("thermo              {}".format(int(nsteps / 100)))

        ncomputes = 0
        ndumps = 0
        nfixes = 0
        if P["thermostat"] == "Nose-Hoover":
            Tchain = P["Tchain"]
            Tloop = P["Tloop"]
            drag = P["drag"]
            if barostat == "Nose-Hoover":
                nfixes += 1
                lines.append(
                    "fix                 {} all npt ".format(nfixes)
                    + "temp {} {} {} ".format(T0, T1, Tdamp)
                    + "tchain {} ".format(Tchain)
                    + "tloop {} ".format(Tloop)
                    + "drag {}".format(drag)
                    + ptext
                )

                for citation in NPT.methods["Nose-Hoover"]["references"]:
                    self.references.cite(
                        raw=self._bibliography[citation],
                        alias=citation,
                        module="lammps_step",
                        level=1,
                        note="Citation for NPT barostat.",
                    )
            else:
                nfixes += 1
                lines.append(
                    "fix                 {} all nvt ".format(nfixes)
                    + "temp {} {} {} ".format(T0, T1, Tdamp)
                    + "tchain {} ".format(Tchain)
                    + "tloop {} ".format(Tloop)
                    + "drag {}".format(drag)
                )
                nfixes += 1
                lines.append(
                    "fix                 {} all ".format(nfixes)
                    + "press/berendsen "
                    + ptext
                    + " modulus {}".format(modulus)
                )

                for citation in NPT.methods["Berendsen"]["references"]:
                    self.references.cite(
                        raw=self._bibliography[citation],
                        alias=citation,
                        module="lammps_step",
                        level=1,
                        note="Citation for NPT barostat.",
                    )
        elif P["thermostat"] == "Berendsen":
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all temp/berendsen "
                + " {} {} {}".format(T0, T1, Tdamp)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        elif "csvr" in P["thermostat"]:
            seed = P["seed"]
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all temp/csvr "
                + " {} {} {} {}".format(T0, T1, Tdamp, seed)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        elif "csld" in P["thermostat"]:
            seed = P["seed"]
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all temp/csld "
                + " {} {} {} {}".format(T0, T1, Tdamp, seed)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        elif P["thermostat"] == "velocity rescaling":
            frequency = P["frequency"]
            nevery = max(1, round(nsteps / (frequency / timestep)))
            window = lammps_step.to_lammps_units(P["window"], quantity="temperature")
            fraction = P["fraction"]
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all temp/rescale "
                + "{} {} {} {} {}".format(nevery, T0, T1, window, fraction)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        elif P["thermostat"] == "Langevin":
            seed = P["seed"]
            nfixes += 1
            lines.append(
                "fix                 {} ".format(nfixes)
                + "all langevin "
                + "{} {} {} {} ".format(T0, T1, Tdamp, seed)
            )
            nfixes += 1
            lines.append("fix                 {} ".format(nfixes) + "all nve")
        else:
            raise RuntimeError(
                "Don't recognize temperature control " + "'{}'".format(P["thermostat"])
            )

        # Add the citation for the thermostat
        metadata = lammps_step.thermostat_metadata
        for citation in metadata[P["thermostat"]]["references"]:
            self.references.cite(
                raw=self._bibliography[citation],
                alias=citation,
                module="lammps_step",
                level=1,
                note="Citation for thermostat.",
            )

        # For the heat flux, if requested, we need extra input
        if P["heat flux"] != "never":
            # Unit conversion factor
            if lammps_step.get_lammps_unit_system() == "metal":
                factor = Q_("eV/Å^2/ps")
            else:
                factor = (
                    Q_("kcal/Å^2/fs/mol") / Q_("kcal/mol") * Q_("kcal/mol").to("kJ")
                )
            factor = factor.m_as("W/m^2")
            lines.append(
                f"""
compute             KE all ke/atom
compute             PE all pe/atom

#          centroid doesn't work with kspace, so split into pair and non-pair parts

compute             S_p all stress/atom NULL pair kspace
compute             S_b all centroid/stress/atom NULL bond angle dihedral improper
compute             flux_p all heat/flux KE PE S_p
compute             flux_b all heat/flux KE PE S_b

#          Conversion from kcal/Å^2/fs/mol to W/m^2")

variable            factor equal {factor}
variable            Jx equal v_factor*(c_flux_p[1]+c_flux_b[1])/vol
variable            Jy equal v_factor*(c_flux_p[2]+c_flux_b[2])/vol
variable            Jz equal v_factor*(c_flux_p[3]+c_flux_b[3])/vol
"""
            )

        # summary output written 10 times during run so we can see progress
        nevery = 10
        nfreq = int(nsteps / 10)
        nrepeat = int(nfreq / nevery)
        nfreq = nevery * nrepeat
        nfixes += 1
        filename = f"@{self._id[-1]}+npt_summary.trj"
        lines.append(
            f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
            f"                       {properties} &\n"
            "                       off 2 &\n"
            f"                       file {filename}"
        )
        # instantaneous output written for averaging
        if P["sampling"] == "none":
            self.description.append(
                __(
                    "The run will be {nsteps:n} steps of dynamics.",
                    nsteps=nsteps,
                    indent=3 * " ",
                )
            )
        else:
            sampling = lammps_step.to_lammps_units(P["sampling"], quantity="time")
            nevery = max(1, round(sampling / timestep))
            nfreq = int(nsteps / nevery)
            nrepeat = 1
            nfreq = nevery * nrepeat
            nfixes += 1
            dt = (nevery * P["timestep"]).to_compact()
            P0tmp = P["Pinitial"].to_compact()
            P1tmp = P["Pfinal"].to(P0tmp.units)
            tmp = {
                "code": "LAMMPS",
                "type": "NPT",
                "dt": dt.magnitude,
                "tunits": str(dt.u),
                "nsteps": nsteps // nevery,
                "Tunits": "K",
                "Punits": str(P0tmp.u),
            }
            if P0tmp == P1tmp:
                tmp["P"] = P0tmp.magnitude
            else:
                tmp["P0"] = P0tmp.magnitude
                tmp["T1"] = P1tmp.magnitude
            if T0 == T1:
                tmp["T"] = T0
            else:
                tmp["T0"] = T0
                tmp["T1"] = T1
            title1 = "!MolSSI trajectory 2.0 " + json.dumps(tmp, separators=(",", ":"))
            filename = f"@{self._id[-1]}+npt_state.trj"
            lines.append(
                f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
                f"                       {properties} &\n"
                "                       off 2 &\n"
                f"                       title1 '{title1}' &\n"
                f"                       title2 '{title2}' &\n"
                f"                       file {filename}"
            )
            self.description.append(
                __(
                    (
                        "The run will be {nsteps:,d} steps of dynamics "
                        "sampled every {nevery:n} steps."
                    ),
                    nsteps=nsteps,
                    nevery=nevery,
                    indent=7 * " ",
                )
            )

        # Handle trajectories
        tmp, ncomputes, ndumps, nfixes = self.trajectory_input(
            P, timestep, nsteps, ncomputes, ndumps, nfixes
        )
        lines.extend(tmp)

        if extras is not None and "shake" in extras:
            nfixes += 1
            lines.append(extras["shake"].format(nfixes))

        lines.append("")
        lines.append("run                 {}".format(nsteps))
        lines.append("")

        for fix in range(1, nfixes + 1):
            lines.append("unfix               {}".format(fix))
        lines.append("")

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }

    def get_pressure_text(self, P, keep_orthorhombic):
        """Work out and return the pressure/stress part of the
        'fix npt' or 'fix berendsen' in LAMMPS
        """
        system_type = P["system type"]
        Panneal = P["Panneal"]
        if system_type == "fluid":
            use_stress = "isotropic pressure"
            couple = "x, y and z"
        else:
            use_stress = P["use_stress"]
            couple = P["couple"]

        if use_stress == "isotropic pressure":
            if couple == "x, y and z":
                if keep_orthorhombic:
                    ptext = " iso {P0} {P1} {Pdamp}"
                else:
                    ptext = (
                        " couple xyz "
                        "x {P0} {P1} {Pdamp} "
                        "y {P0} {P1} {Pdamp} "
                        "z {P0} {P1} {Pdamp}"
                    )
                if not keep_orthorhombic:
                    ptext += (
                        " xy 0.0 0.0 {Pdamp} "
                        "xz 0.0 0.0 {Pdamp} "
                        "yz 0.0 0.0 {Pdamp}"
                    )
            elif couple == "x and y":
                ptext = (
                    " couple xy "
                    "x {P0} {P1} {Pdamp} "
                    "y {P0} {P1} {Pdamp} "
                    "z {P0} {P1} {Pdamp}"
                )
                if not keep_orthorhombic:
                    ptext += (
                        " xy 0.0 0.0 {Pdamp} "
                        "xz 0.0 0.0 {Pdamp} "
                        "yz 0.0 0.0 {Pdamp}"
                    )
            elif couple == "x and z":
                ptext = (
                    " couple xz "
                    "x {P0} {P1} {Pdamp} "
                    "y {P0} {P1} {Pdamp} "
                    "z {P0} {P1} {Pdamp}"
                )
                if not keep_orthorhombic:
                    ptext += (
                        " xy 0.0 0.0 {Pdamp} "
                        "xz 0.0 0.0 {Pdamp} "
                        "yz 0.0 0.0 {Pdamp}"
                    )
            elif couple == "y and z":
                ptext = (
                    " couple yz "
                    "x {P0} {P1} {Pdamp} "
                    "y {P0} {P1} {Pdamp} "
                    "z {P0} {P1} {Pdamp}"
                )
                if not keep_orthorhombic:
                    ptext += (
                        " xy 0.0 0.0 {Pdamp} "
                        "xz 0.0 0.0 {Pdamp} "
                        "yz 0.0 0.0 {Pdamp}"
                    )
            else:
                if keep_orthorhombic:
                    ptext = " aniso {P0} {P1} {Pdamp}"
                else:
                    ptext = " tri {P0} {P1} {Pdamp}"

            P0 = lammps_step.to_lammps_units(P["Pinitial"], quantity="pressure")
            if Panneal:
                P1 = lammps_step.to_lammps_units(P["Pfinal"], quantity="pressure")
            else:
                P1 = P0
            Pdamp = lammps_step.to_lammps_units(P["Pdamp"], quantity="time")

            ptext = ptext.format(P0=P0, P1=P1, Pdamp=Pdamp)
        else:
            if couple == "x, y and z":
                ptext = (
                    " couple = xyz "
                    "x {Sxx0} {Sxx1} {Dxx} "
                    "y {Sxx0} {Sxx1} {Dxx} "
                    "z {Sxx0} {Sxx1} {Dxx}"
                )
            elif couple == "x and y":
                ptext = (
                    " couple = xy "
                    "x {Sxx0} {Sxx1} {Dxx} "
                    "y {Sxx0} {Sxx1} {Dxx} "
                    "z {Szz0} {Szz1} {Dzz}"
                )
            elif couple == "x and z":
                ptext = (
                    " couple = xz "
                    "x {Sxx0} {Sxx1} {Dxx} "
                    "y {Syy0} {Syy1} {Dyy} "
                    "z {Sxx0} {Sxx1} {Dxx}"
                )
            elif couple == "y and z":
                ptext = (
                    " couple = yz "
                    "x {Sxx0} {Sxx1} {Dxx} "
                    "y {Syy0} {Syy1} {Dyy} "
                    "z {Syy0} {Syy1} {Dyy}"
                )
            else:
                # elif couple == 'none':
                ptext = (
                    " couple = none "
                    "x {Sxx0} {Sxx1} {Dxx} "
                    "y {Syy0} {Syy1} {Dyy} "
                    "z {Szz0} {Szz1} {Dzz}"
                )

            if not keep_orthorhombic:
                ptext += (
                    " xy {Sxy0} {Sxy1} {Dxy} "
                    "xz {Sxz0} {Sxz1} {Dxz} "
                    "yz {Syz0} {Syz1} {Dyz}"
                )

            Tmp = {}
            Tmp["Sxx0"] = lammps_step.to_lammps_units(
                P["Sxx,initial"], quantity="pressure"
            )
            Tmp["Syy0"] = lammps_step.to_lammps_units(
                P["Syy,initial"], quantity="pressure"
            )
            Tmp["Szz0"] = lammps_step.to_lammps_units(
                P["Szz,initial"], quantity="pressure"
            )
            if Panneal:
                Tmp["Sxx1"] = lammps_step.to_lammps_units(
                    P["Sxx,final"], quantity="pressure"
                )
                Tmp["Syy1"] = lammps_step.to_lammps_units(
                    P["Syy,final"], quantity="pressure"
                )
                Tmp["Szz1"] = lammps_step.to_lammps_units(
                    P["Szz,final"], quantity="pressure"
                )
            else:
                Tmp["Sxx1"] = Tmp["Sxx0"]
                Tmp["Syy1"] = Tmp["Syy0"]
                Tmp["Szz1"] = Tmp["Szz0"]
            Tmp["Dxx"] = lammps_step.to_lammps_units(P["Sxx damp"], quantity="pressure")
            Tmp["Dyy"] = lammps_step.to_lammps_units(P["Syy damp"], quantity="pressure")
            Tmp["Dzz"] = lammps_step.to_lammps_units(P["Szz damp"], quantity="pressure")

            if not keep_orthorhombic:
                Tmp["Sxy0"] = lammps_step.to_lammps_units(
                    P["Sxy,initial"], quantity="pressure"
                )
                Tmp["Sxz0"] = lammps_step.to_lammps_units(
                    P["Sxz,initial"], quantity="pressure"
                )
                Tmp["Syz0"] = lammps_step.to_lammps_units(
                    P["Syz,initial"], quantity="pressure"
                )
                if Panneal:
                    Tmp["Sxy1"] = lammps_step.to_lammps_units(
                        P["Sxy,final"], quantity="pressure"
                    )
                    Tmp["Sxz1"] = lammps_step.to_lammps_units(
                        P["Sxz,final"], quantity="pressure"
                    )
                    Tmp["Syz1"] = lammps_step.to_lammps_units(
                        P["Syz,final"], quantity="pressure"
                    )
                else:
                    Tmp["Sxy1"] = Tmp["Sxy0"]
                    Tmp["Sxz1"] = Tmp["Sxz0"]
                    Tmp["Syz1"] = Tmp["Syz0"]
                Tmp["Dxy"] = lammps_step.to_lammps_units(
                    P["Sxy damp"], quantity="pressure"
                )
                Tmp["Dxz"] = lammps_step.to_lammps_units(
                    P["Sxz damp"], quantity="pressure"
                )
                Tmp["Dyz"] = lammps_step.to_lammps_units(
                    P["Syz damp"], quantity="pressure"
                )

            ptext = ptext.format(**Tmp)

        return ptext



# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS minimization step"""

import lammps_step
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk


class TkMinimization(lammps_step.TkEnergy):
    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=None, y=None, w=200, h=50
    ):
        """Initialize a node

        Keyword arguments:
        """

        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

    def create_dialog(self):
        """Create the dialog!"""
        super().create_dialog()
        self.dialog.configure(title="Edit Minimization step")

        frame = self["frame"]

        # Convergence
        convergence_label = ttk.Label(frame, text="Simulation convergence:")
        self["convergence_label"] = convergence_label

        convergence = ttk.Combobox(
            frame,
            state="readonly",
            values=[
                "normal",
                "tight",
                "loose",
                "crude",
                "on forces",
                "on energy and forces",
                "on just the energy change",
                "from variable",
            ],
            justify=tk.LEFT,
            width=20,
        )
        convergence.set(self.node.convergence)
        self["convergence"] = convergence

        # etol: a variable or a value...
        etol_label = ttk.Label(frame, text="change in energy:")
        self["etol_label"] = etol_label

        etol_method = ttk.Combobox(
            frame,
            state="readonly",
            values=[
                "is",
                "from variable",
            ],
            justify=tk.LEFT,
            width=10,
        )
        etol_method.set(self.node.etol_method)
        self["etol_method"] = etol_method

        etol_variable = ttk.Entry(frame, width=15)
        etol_variable.insert(0, self.node.etol_variable)
        self["etol_variable"] = etol_variable

        etol = ttk.Entry(frame, width=15)
        etol.insert(0, self.node.etol)
        self["etol"] = etol

        # ftol: a variable or a value...
        ftol_label = ttk.Label(frame, text="magnitude of the force vector:")
        self["ftol_label"] = ftol_label

        ftol_method = ttk.Combobox(
            frame,
            state="readonly",
            values=[
                "is",
                "from variable",
            ],
            justify=tk.LEFT,
            width=10,
        )
        ftol_method.set(self.node.ftol_method)
        self["ftol_method"] = ftol_method

        ftol_variable = ttk.Entry(frame, width=15)
        ftol_variable.insert(0, self.node.ftol_variable)
        self["ftol_variable"] = ftol_variable

        ftol = sw.UnitEntry(frame, width=15)
        ftol.set(self.node.ftol)
        self["ftol"] = ftol

        # maxiters: a variable or a value...
        maxiters_label = ttk.Label(frame, text="maximum iterations:")
        self["maxiters_label"] = maxiters_label

        maxiters_method = ttk.Combobox(
            frame,
            state="readonly",
            values=[
                "default",
                "is",
                "from variable",
            ],
            justify=tk.LEFT,
            width=10,
        )
        maxiters_method.set(self.node.maxiters_method)
        self["maxiters_method"] = maxiters_method

        maxiters_variable = ttk.Entry(frame, width=15)
        maxiters_variable.insert(0, self.node.maxiters_variable)
        self["maxiters_variable"] = maxiters_variable

        maxiters = ttk.Entry(frame, width=15)
        maxiters.insert(0, self.node.maxiters)
        self["maxiters"] = maxiters

        # maxevals: a variable or a value...
        maxevals_label = ttk.Label(frame, text="maximum energy evaluations:")
        self["maxevals_label"] = maxevals_label

        maxevals_method = ttk.Combobox(
            frame,
            state="readonly",
            values=[
                "default",
                "is",
                "from variable",
            ],
            justify=tk.LEFT,
            width=10,
        )
        maxevals_method.set(self.node.maxevals_method)
        self["maxevals_method"] = maxevals_method

        maxevals_variable = ttk.Entry(frame, width=15)
        maxevals_variable.insert(0, self.node.maxevals_variable)
        self["maxevals_variable"] = maxevals_variable

        maxevals = ttk.Entry(frame, width=15)
        maxevals.insert(0, self.node.maxevals)
        self["maxevals"] = maxevals

        convergence.bind("<<ComboboxSelected>>", self.reset_dialog)
        etol_method.bind("<<ComboboxSelected>>", self.reset_dialog)
        ftol_method.bind("<<ComboboxSelected>>", self.reset_dialog)
        maxiters_method.bind("<<ComboboxSelected>>", self.reset_dialog)
        maxevals_method.bind("<<ComboboxSelected>>", self.reset_dialog)

    def reset_dialog(self, widget=None):
        """Layout the widgets as needed for the current state"""
        convergence = self["convergence"].get()
        etol_method = self["etol_method"].get()
        ftol_method = self["ftol_method"].get()
        maxiters_method = self["maxiters_method"].get()
        maxevals_method = self["maxevals_method"].get()

        frame = self["frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        row = 0
        self["convergence_label"].grid(row=row, column=0, sticky=tk.E)
        self["convergence"].grid(row=row, column=1, columnspan=2, sticky=tk.W)

        if "energy" in convergence:
            row += 1
            self["etol_label"].grid(row=row, column=0, sticky=tk.E)
            self["etol_method"].grid(row=row, column=1, sticky=tk.EW)
            if etol_method == "from variable":
                self["etol_variable"].grid(row=row, column=2, sticky=tk.W)
            elif etol_method == "is":
                self["etol"].grid(row=row, column=2, sticky=tk.W)

        if "forces" in convergence:
            row += 1
            self["ftol_label"].grid(row=row, column=0, sticky=tk.E)
            self["ftol_method"].grid(row=row, column=1, sticky=tk.EW)
            if ftol_method == "from variable":
                self["ftol_variable"].grid(row=row, column=2, sticky=tk.W)
            elif ftol_method == "is":
                self["ftol"].grid(row=row, column=2, sticky=tk.W)

        if "energy" in convergence or "forces" in convergence:
            row += 1
            self["maxiters_label"].grid(row=row, column=0, sticky=tk.E)
            self["maxiters_method"].grid(row=row, column=1, sticky=tk.EW)
            if maxiters_method == "from variable":
                self["maxiters_variable"].grid(row=row, column=2, sticky=tk.W)
            elif maxiters_method == "is":
                self["maxiters"].grid(row=row, column=2, sticky=tk.W)

            row += 1
            self["maxevals_label"].grid(row=row, column=0, sticky=tk.E)
            self["maxevals_method"].grid(row=row, column=1, sticky=tk.EW)
            if maxevals_method == "from variable":
                self["maxevals_variable"].grid(row=row, column=2, sticky=tk.W)
            elif maxevals_method == "is":
                self["maxevals"].grid(row=row, column=2, sticky=tk.W)

    def handle_dialog(self, result):
        if result is None or result == "Cancel":
            self.dialog.deactivate(result)
            return

        if result == "Help":
            # display help!!!
            return

        if result != "OK":
            self.dialog.deactivate(result)
            raise RuntimeError("Don't recognize dialog result '{}'".format(result))

        self.dialog.deactivate(result)

        convergence = self["convergence"].get()
        etol_method = self["etol_method"].get()
        ftol_method = self["ftol_method"].get()
        maxiters_method = self["maxiters_method"].get()
        maxevals_method = self["maxevals_method"].get()

        self.node.convergence = convergence
        if "energy" in convergence:
            self.node.etol_method = etol_method
            if etol_method == "from variable":
                self.node.etol_variable = self["etol_variable"].get()
            elif etol_method == "is":
                self.node.etol = int(self["etol"].get())

        if "forces" in convergence:
            self.node.ftol_method = ftol_method
            if ftol_method == "from variable":
                self.node.ftol_variable = self["ftol_variable"].get()
            elif ftol_method == "is":
                self.node.ftol = self["ftol"].get()

        if "energy" in convergence or "forces" in convergence:
            self.node.maxiters_method = maxiters_method
            if maxiters_method == "from variable":
                self.node.maxtiters_variable = self["maxiters_variable"].get()
            elif maxiters_method == "is":
                self.node.maxiters = int(self["maxiters"].get())

            self.node.maxevals_method = maxevals_method
            if maxevals_method == "from variable":
                self.node.maxevals_variable = self["maxevals_variable"].get()
            elif maxevals_method == "is":
                self.node.maxevals = int(self["maxevals"].get())



# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS Energy step"""

import lammps_step
import logging
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk

logger = logging.getLogger(__name__)


class TkNVE(lammps_step.TkEnergy):
    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """Initialize a node

        Keyword arguments:
        """

        # Metadata for the properties
        self.property_metadata = {}

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def create_dialog(self, title="Edit NVE dynamics parameters"):
        """Create the dialog!"""

        # Let parent classes do their thing.
        frame = super().create_dialog(title=title)

        # Shortcut for parameters
        P = self.node.parameters

        # Add a tab for the trajectory setup
        notebook = self["notebook"]
        tframe = ttk.Frame(notebook)
        self["trajectory frame"] = tframe
        notebook.insert(
            self["results frame"], tframe, text="Trajectories", sticky=tk.NSEW
        )

        row = 0
        widgets = []
        for key in lammps_step.NVE_Parameters.trajectories:
            if title == "Heat Flux" and (key == "heat flux" or "centroid" in key):
                continue
            self[key] = P[key].widget(tframe)
            self[key].grid(row=row, column=0)
            row += 1
            widgets.append(self[key])
        sw.align_labels(widgets, sticky=tk.E)

        if title == "Heat Flux":
            return frame

        # Frame to isolate widgets
        c_frame = self["control_frame"] = ttk.LabelFrame(
            self["frame"],
            borderwidth=4,
            relief="sunken",
            text="General Parameters",
            labelanchor="n",
            padding=10,
        )

        for key in lammps_step.NVE_Parameters.parameters:
            if key == "control_properties":
                self[key] = P[key].widget(c_frame, metadata=self.property_metadata)
            else:
                self[key] = P[key].widget(c_frame)

        # make the control combobox wide enough
        self["run_control"].combobox.configure(width=40)

        # and binding to change as needed
        self["run_control"].combobox.bind(
            "<<ComboboxSelected>>", self.reset_control_frame
        )

        return frame

    def reset_dialog(self, widget=None):
        """Layout the widgets as needed for the current state"""

        frame = self["frame"]
        # Clear the dialog
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # Put in our control frame
        row = 0
        self["control_frame"].grid(row=row, column=0)
        row += 1

        # and the widgets in it
        self.reset_control_frame()

        return row

    def reset_control_frame(self, widget=None):
        """Layout the control widgets as needed for the current state"""

        run_control = self["run_control"].get()

        # Clear out the previous widgets
        c_frame = self["control_frame"]
        for slave in c_frame.grid_slaves():
            slave.grid_forget()

        # And put them back in depending...
        row = 0
        # self["run_control"].grid(row=row, column=0, sticky=tk.W)
        # row += 1

        widgets = []
        if "fixed length" in run_control:
            widgets.append(self["time"])
            self["time"].grid(row=row, column=0, sticky=tk.W)
            row += 1
        else:
            widgets.append(self["maximum_time"])
            self["maximum_time"].grid(row=row, column=0, sticky=tk.W)
            row += 1
        widgets.append(self["timestep"])
        self["timestep"].grid(row=row, column=0, sticky=tk.W)
        row += 1
        if "fixed length" in run_control:
            widgets.append(self["sampling"])
            self["sampling"].grid(row=row, column=0, sticky=tk.W)
            row += 1
        else:
            self["control_properties"].grid(row=row, column=0, sticky=tk.NSEW)
            row += 1

        sw.align_labels(widgets, sticky=tk.E)

    def handle_dialog(self, result):
        if result == "OK":
            # Shortcut for parameters
            P = self.node.parameters

            value, units = self["time"].get()
            P["time"].value = value
            P["time"].units = units

            tmp = self["timestep"].get()
            if tmp in P["timestep"].enumeration:
                P["timestep"].value = tmp
            else:
                P["timestep"].value = tmp[0]
                P["timestep"].units = tmp[1]

            tmp = self["sampling"].get()
            if tmp in P["sampling"].enumeration:
                P["sampling"].value = tmp
            else:
                P["sampling"].value = tmp[0]
                P["sampling"].units = tmp[1]

        # Let base classes reap their parameters
        super().handle_dialog(result)



# -*- coding: utf-8 -*-

"""A single-point initialization in LAMMPS"""

import lammps_step
import logging
import seamm
import seamm_util
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import pprint

logger = logging.getLogger("lammps")
job = printing.getPrinter()
printer = printing.getPrinter("lammps")

msm_pair_styles = ["born", "buck", "", "lj/charmm", "lj/cut"]

thermo_variables = [
    "step",
    "elapsed",
    "elaplong",
    "dt",
    "time",
    "cpu",
    "tpcpu",
    "spcpu",
    "cpuremain",
    "part",
    "timeremain",
    "atoms",
    "temp",
    "press",
    "pe",
    "ke",
    "etotal",
    "enthalpy",
    "evdwl",
    "ecoul",
    "epair",
    "ebond",
    "eangle",
    "edihed",
    "eimp",
    "emol",
    "elong",
    "etail",
    "vol",
    "density",
    "lx",
    "ly",
    "lz",
    "xlo",
    "xhi",
    "ylo",
    "yhi",
    "zlo",
    "zhi",
    "xy",
    "xz",
    "yz",
    "xlat",
    "ylat",
    "zlat",
    "bonds",
    "angles",
    "dihedrals",
    "impropers",
    "pxx",
    "pyy",
    "pzz",
    "pxy",
    "pxz",
    "pyz",
    "fmax",
    "fnorm",
    "nbuild",
    "ndanger",
    "cella",
    "cellb",
    "cellc",
    "cellalpha",
    "cellbeta",
    "cellgamma",
]


class Initialization(seamm.Node):
    def __init__(self, flowchart=None, title="Initialization", extension=None):
        """Initialize the node"""

        logger.debug("Creating Initialization {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self.description = []
        self.parameters = lammps_step.InitializationParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return lammps_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return lammps_step.__git_revision__

    @property
    def kspace_methods(self):
        """The list of avilable methods"""
        return list(lammps_step.kspace_methods)

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """

        if not P:
            P = self.parameters.values_to_dict()

        text = "Initialize the calculation with a cutoff of {cutoff}"
        if P["shift_nonbond"]:
            text += ", shifting the nonbond energies to 0 at the cutoff"
        text += ". If the system is periodic"
        if P["kspace_method"][0] == "$":
            text += " use the variable '{method}' to determine whether "
            text += "and how to accelerate the k-space summations."
        elif P["kspace_method"] == "none":
            text += " no k-space acceleration method will be used."
        elif P["kspace_method"] == "automatic":
            text += " the best k-space acceleration method for the "
            text += " molecular system will be chosen."
        else:
            text += " the {method} for k-space acceleration will be used."

        if P["kspace_method"] != "none":
            text += " The accuracy goal is {kspace_accuracy}."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self, extras=None):
        """Get the input for the initialization of LAMMPS"""

        self.description = []
        self.description.append("   " + self.header)

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Fix some things
        P["cutoff"] = P["cutoff"].to("angstrom").magnitude

        # Get the system
        system_db = self.get_variable("_system_db")
        configuration = system_db.system.configuration

        # See what type of forcefield we have and handle it
        ff = self.get_variable("_forcefield")
        if ff == "OpenKIM":
            lammps_step.set_lammps_unit_system("metal")
            return self.OpenKIM_input()

        # Valence forcefield...
        ffname = ff.current_forcefield
        n_atoms = configuration.n_atoms

        # And atom-type if necessary
        key = f"atom_types_{ffname}"
        if key not in configuration.atoms:
            logger.debug("Atom typing")
            ff.assign_forcefield(configuration)
        else:
            if any(typ is None for typ in configuration.atoms[key]):
                ff.assign_forcefield(configuration)

        # Get the energy expression.
        style = "LAMMPS-class2" if "cff" in ffname else "LAMMPS"
        eex = ff.energy_expression(configuration, style=style)
        logger.debug("energy expression:\n" + pprint.pformat(eex))

        # Determine if we have any charges, and if so, if they are sparse
        key = f"charges_{ffname}"
        if key in configuration.atoms:
            charges = [*configuration.atoms[key]]
            n_charged_atoms = 0
            smallq = float(P["kspace_smallq"])
            for charge in charges:
                if abs(charge) > smallq:
                    n_charged_atoms += 1
            fraction_charged_atoms = n_charged_atoms / n_atoms
        else:
            n_charged_atoms = 0

        kspace_style = ""

        lines = []
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append("units               real")

        periodicity = configuration.periodicity
        if periodicity == 0:
            lines.append("boundary            s s s")
            string = "Setup for a molecular (non-periodic) system."
        elif periodicity == 3:
            lines.append("boundary            p p p")
            tail_correction = (
                "yes" if P["tail_correction"] and not P["shift_nonbond"] else "no"
            )
            string = "Setup for a periodic (crystalline or fluid) system."
        else:
            raise RuntimeError(
                "The LAMMPS step can only handle 0-"
                " or 3-D periodicity at the moment!"
            )

        lines.append("atom_style          full")
        lines.append("newton              on")
        if n_atoms < 20:
            # LAMMPS has problems with bins for small systems
            lines.append("neighbor            2.0 nsq")
        lines.append("")
        if periodicity == 3:
            a, b, c, alpha, beta, gamma = configuration.cell.parameters
            if alpha != 90 or beta != 90 or gamma != 90:
                lines.append("box                 tilt large")
        lines.append("#    define the style of forcefield")
        lines.append("")

        terms = ff.ff["terms"]

        logging.debug("LAMMPS initialization, terms = \n" + pprint.pformat(terms))

        # control of nonbonds
        nonbond_term = None
        if "pair" in terms:
            if len(terms["pair"]) != 1:
                raise RuntimeError("Cannot handle multiple nonbond terms yet!")
            nonbond_term = terms["pair"][0]

        if nonbond_term is None:
            pprint.pprint(terms)
            raise RuntimeError("Cannot find nonbond term in forcefield!")

        if nonbond_term == "nonbond(9-6)":
            pair_style_base = "lj/class2"
            mixing = "sixthpower"
        elif nonbond_term == "nonbond(12-6)":
            pair_style_base = "lj/cut"
            # What type of mixing rule?
            modifiers = ff.ff["modifiers"]["nonbond(12-6)"]
            mixing = ""
            for section in modifiers:
                for item in modifiers[section]:
                    if "combination" in item:
                        if mixing == "":
                            mixing = item.split()[1]
                        if mixing != item.split()[1]:
                            raise RuntimeError(
                                "Conflicting combination rules in "
                                "nonbond(12-6) section '" + section + "'"
                            )
            if mixing == "":
                mixing = "geometric"
        elif nonbond_term == "buckingham":
            pair_style_base = "buck"
            mixing = None
        else:
            raise RuntimeError("Can't handle nonbond term {} yet!".format(nonbond_term))

        shift = "yes" if P["shift_nonbond"] else "no"
        if P["kspace_method"] == "automatic":
            if periodicity == 3:
                if n_charged_atoms == 0:
                    pair_style = pair_style_base
                    string += (
                        " The nonbonded interactions will be evaluated using "
                        "a cutoff of {cutoff} Å. Since there are no charges "
                        "on the atoms, no long-range coulomb method will be "
                        "used."
                    )
                else:
                    string += (
                        " The nonbonded interactions will be evaluated using "
                        "a cutoff of {cutoff} Å, with the long-range terms "
                    )
                    pair_style = pair_style_base + "/coul/long"
                    if n_atoms < P["ewald_atom_cutoff"]:
                        kspace_style = "ewald {}".format(P["kspace_accuracy"])
                        string += (
                            "using the Ewald summation method with "
                            "an accuracy of {kspace_accuracy}."
                        )
                    elif fraction_charged_atoms < P["charged_atom_fraction_cutoff"]:
                        kspace_style = "pppm/cg {} {}".format(
                            P["kspace_accuracy"], P["charged_atom_fraction_cutoff"]
                        )
                        string += (
                            "using the PPPM method optimized for few "
                            "atoms with charges, with "
                            "an accuracy of {kspace_accuracy}."
                        )
                    else:
                        kspace_style = "pppm {}".format(P["kspace_accuracy"])
                        string += (
                            "using the PPPM method with "
                            "an accuracy of {kspace_accuracy}."
                        )
                lines.append(
                    "pair_style          {} {}".format(pair_style, P["cutoff"])
                )
                if mixing is None:
                    lines.append(
                        f"pair_modify         tail {tail_correction} shift {shift}"
                    )
                else:
                    lines.append(
                        "pair_modify         mix "
                        + mixing
                        + " tail {} shift {}".format(tail_correction, shift)
                    )
                if shift:
                    string += (
                        " The van der Waals terms will be shifted "
                        "to zero energy at the cutoff distance."
                    )
                if tail_correction:
                    string += (
                        " A long-range correction for the "
                        "van der Waals terms will be added."
                    )
                if kspace_style != "":
                    lines.append("kspace_style        " + kspace_style)
            else:
                kspace_style = ""
                if n_charged_atoms == 0:
                    pair_style = pair_style_base
                    string += (
                        " The nonbonded interactions will be evaluated using "
                        "a simple cutoff of {cutoff} Å. Since there are no "
                        "charges on the atoms, no long-range coulomb method "
                        "will be used."
                    )
                elif (
                    n_atoms > P["msm_atom_cutoff"]
                    and pair_style_base in msm_pair_styles
                ):
                    pair_style = pair_style_base + "/coul/msm"
                    string += (
                        "The nonbonded interactions will be handled with "
                        " a cutoff of {cutoff} Å."
                    )
                    if fraction_charged_atoms < P["charged_atom_fraction_cutoff"]:
                        kspace_style = "msm/cg {kspace_accuracy} {kspace_smallq}"
                        string += (
                            " The MSM method will be used to handle "
                            "the longer range coulombic interactions, using "
                            "the approach tuned for systems with few charges."
                            "The accuracy goal is {kspace_accuracy}."
                        )
                    else:
                        kspace_style = "msm {kspace_accuracy}"
                        string += (
                            " The MSM method will be used to handle "
                            "the longer range coulombic interactions."
                            "The accuracy goal is {kspace_accuracy}."
                        )
                else:
                    pair_style = pair_style_base + "/coul/cut"
                    string += (
                        "The nonbonded interactions will be handled with "
                        " a simple cutoff of {cutoff} Å."
                    )
                if shift:
                    string += (
                        " The van der Waals terms will be shifted "
                        "to zero energy at the cutoff distance."
                    )
                lines.append(
                    "pair_style          {} {}".format(pair_style, P["cutoff"])
                )
                if mixing is None:
                    lines.append(f"pair_modify         shift {shift}")
                else:
                    lines.append(f"pair_modify         mix {mixing} shift {shift}")
                if kspace_style != "":
                    lines.append("kspace_style        " + kspace_style.format(**P))
            self.description.append(__(string, indent=7 * " ", **P))
        else:
            if periodicity == 3:
                kspace_style = ""
                if n_charged_atoms == 0 or P["kspace_style"] == "none":
                    pair_style = pair_style_base
                elif fraction_charged_atoms < P["charged_atom_fraction_cutoff"]:
                    pair_style = pair_style_base + "/coul/long"
                    kspace_style = lammps_step.kspace_methods[
                        P["kspace_method"]
                    ].format(**P)
                lines.append(
                    "pair_style          {} {}".format(pair_style, P["cutoff"])
                )
                if mixing is None:
                    lines.append(
                        f"pair_modify         tail {tail_correction} shift {shift}"
                    )
                else:
                    lines.append(
                        "pair_modify         mix "
                        + mixing
                        + " tail {} shift {}".format(tail_correction, shift)
                    )
            else:
                if n_charged_atoms == 0:
                    pair_style = pair_style_base
                else:
                    pair_style = pair_style_base + "/coul/cut"
                lines.append(
                    "pair_style          {} {}".format(pair_style, P["cutoff"])
                )
                if mixing is None:
                    lines.append(f"pair_modify         shift {shift}")
                else:
                    lines.append(f"pair_modify         mix {mixing} shift {shift}")
                if "msm" in lammps_step.kspace_methods[P["kspace_method"]]:
                    kspace_style = lammps_step.kspace_methods[
                        P["kspace_method"]
                    ].format(**P)

        if "n_bonds" in eex and eex["n_bonds"] > 0:
            forms = set([v[0] for v in eex["bond parameters"]])
            if len(forms) == 1:
                bond_style = lammps_step.bond_style[[*forms][0]]
                lines.append("bond_style          " + bond_style)
            else:
                line = "bond_style          hybrid"
                for term in forms:
                    line += " " + lammps_step.bond_style[term]
                lines.append(line)
        if "angle" in terms and eex["n_angles"] > 0:
            forms = set([v[0] for v in eex["angle parameters"]])
            if len(forms) == 1:
                angle_style = lammps_step.angle_style[[*forms][0]]
                if angle_style == "table":
                    lines.append(f"angle_style         {angle_style} linear 901")
                else:
                    lines.append("angle_style         " + angle_style)
            else:
                line = "angle_style         hybrid"
                for term in forms:
                    angle_style = lammps_step.angle_style[term]
                    if angle_style == "table":
                        line += f" {angle_style} linear 901"
                    else:
                        line += " " + angle_style
                lines.append(line)
        if "torsion" in terms and eex["n_torsions"] > 0:
            forms = set([v[0] for v in eex["torsion parameters"]])
            if len(forms) == 1:
                dihedral_style = lammps_step.dihedral_style[[*forms][0]]
                lines.append("dihedral_style      " + dihedral_style)
            else:
                line = "dihedral_style      hybrid"
                for term in forms:
                    line += " " + lammps_step.dihedral_style[term]
                lines.append(line)
        if "out-of-plane" in terms and eex["n_oops"] > 0:
            forms = set([v[0] for v in eex["oop parameters"]])
            if len(forms) == 1:
                improper_style = lammps_step.improper_style[[*forms][0]]
                lines.append("improper_style      " + improper_style)
            else:
                line = "improper_style      hybrid"
                for term in forms:
                    line += " " + lammps_step.improper_style[term]
                lines.append(line)

        lines.append("")

        # Handle 1-4s in nonbonds
        if "opls" in ffname:
            lines.append("special_bonds       lj/coul 0.0 0.0 0.5")
            lines.append("")

        if extras is not None and "read_data" in extras and extras["read_data"] is True:
            lines.append("read_data           structure.dat")

        if kspace_style != "":
            lines.append("kspace_style        " + kspace_style)

        # Set up standard variables
        for variable in thermo_variables:
            lines.append("variable            {var} equal {var}".format(var=variable))

        return (lines, eex)

    def OpenKIM_input(self):
        """Create the initialization input for a calculation using OpenKIM."""
        # Get the configuration
        system_db = self.get_variable("_system_db")
        configuration = system_db.system.configuration

        # Get the (simple) energy expression for these systems
        eex = self.OpenKIM_energy_expression()

        lines = []
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append("newton              on")
        lines.append("")
        potential = self.get_variable("_OpenKIM_Potential")
        if "reax" in potential.lower():
            lines.append("atom_style          charge")
            lines.append(
                f"kim                 init {potential} real unit_conversion_mode"
            )
        else:
            lines.append(
                f"kim                 init {potential} metal unit_conversion_mode"
            )
        lines.append("")
        periodicity = configuration.periodicity
        if periodicity == 0:
            lines.append("boundary            s s s")
            string = "Setup for a molecular (non-periodic) system."
        elif periodicity == 3:
            lines.append("boundary            p p p")
            string = "Setup for a periodic (crystalline or fluid) system."
        else:
            raise RuntimeError(
                "The LAMMPS step can only handle 0-"
                " or 3-D periodicity at the moment!"
            )
        lines.append("")
        lines.append("read_data           structure.dat")
        lines.append("")
        lines.append(f'kim                 interactions {" ".join(eex["atom types"])}')

        # Set up standard variables
        for variable in thermo_variables:
            lines.append("variable            {var} equal {var}".format(var=variable))

        self.description.append(__(string, indent=self.indent + 4 * " "))

        return (lines, eex)

    def OpenKIM_energy_expression(self):
        """Create the (simple) energy expression for OpenKIM models."""
        eex = {}
        eex["terms"] = {"OpenKIM": []}

        # Get the configuration
        system_db = self.get_variable("_system_db")
        configuration = system_db.system.configuration
        atoms = configuration.atoms

        # The elements (1-based!) Probably not used...
        elements = atoms.symbols
        eex["elements"] = [""]
        eex["elements"].extend(elements)

        # The periodicity & cell parameters
        periodicity = eex["periodicity"] = configuration.periodicity
        if periodicity == 3:
            eex["cell"] = configuration.cell.parameters

        result = eex["atoms"] = []
        atom_types = eex["atom types"] = []
        masses = eex["masses"] = []

        coordinates = atoms.get_coordinates(fractionals=False)
        for element, xyz in zip(elements, coordinates):
            if element in atom_types:
                index = atom_types.index(element) + 1
            else:
                atom_types.append(element)
                index = len(atom_types)
                masses.append(
                    (seamm_util.element_data[element]["atomic weight"], element)
                )
            x, y, z = xyz
            result.append((x, y, z, index))

        eex["n_atoms"] = n_atoms = len(result)
        eex["n_atom_types"] = len(atom_types)

        # For ReaxFF we need charges
        potential = self.get_variable("_OpenKIM_Potential")
        if "reax" in potential.lower():
            if "charge" in atoms:
                eex["charges"] = atoms.get_column_data["charges"]
            else:
                eex["charges"] = [0.0] * n_atoms

        return eex



# -*- coding: utf-8 -*-

"""Installer for the LAMMPS plug-in.

This handles any further installation needed after installing the Python
package `lammps-step`.
"""

import logging
import os
from pathlib import Path
import pkg_resources
import platform
import subprocess

import seamm_installer

logger = logging.getLogger(__name__)


class Installer(seamm_installer.InstallerBase):
    """Handle further installation needed after installing lammps-step.

    The Python package `lammps-step` should already be installed, using `pip`,
    `conda`, or similar. This plug-in-specific installer then checks for the
    LAMMPS executables, installing them if need be, and registers their
    location in seamm.ini.

    .. note:: The LAMMPS step and this assume that the LAMMPS
       executables, if present, are in the same directory. In
       addition, for the parallel version using MPI it is assumed that
       `mpiexec` is either in that same directory, or the correct
       `mpiexec` can be found in the PATH.

    There are a number of ways to determine which are the correct LAMMPS
    executables to use. The aim of this installer is to help the user locate
    the executables. There are a number of possibilities

    #. The correct executables are already available.

        #. If they are already registered in `seamm.ini` there is nothing else
           to do.

        #. They may be in the current path, in which case they need to be added
           to `seamm.ini`.

        #. If a module system is in use, a module may need to be loaded to give
           access to LAMMPS.

        #. They cannot be found automatically, so the user needs to locate the
           executables for the installer.

    #. LAMMPS is not installed on the machine. In this case they can be
       installed in a Conda environment. There is one choice

        #. They can be installed in a separate environment, `seamm-lammps` by
           default.
    """

    def __init__(self, logger=logger):
        # Call the base class initialization, which sets up the commandline
        # parser, amongst other things.
        super().__init__(logger=logger)

        logger.debug("Initializing the LAMMPS installer object.")

        self.section = "lammps-step"
        self.executables = ["lmp"]
        self.resource_path = Path(pkg_resources.resource_filename(__name__, "data/"))

        # The environment.yaml file for Conda installations.
        logger.debug(f"data directory: {self.resource_path}")
        self.environment_file = self.resource_path / "seamm-lammps.yml"

    def check(self):
        """Check the status of the LAMMPS installation."""
        print("Checking the LAMMPS installation.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "lammps.ini"
        if not path.exists():
            text = (self.resource_path / "lammps.ini").read_text()
            path.write_text(text)
            print(f"    The lammps.ini file did not exist. Created {path}")

        self.exe_config.path = path

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-lammps"

        super().check()

    def install(self):
        """Install LAMMPS in a conda environment."""
        print("Installing LAMMPS.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "lammps.ini"
        if not path.exists():
            text = (self.resource_path / "lammps.ini").read_text()
            path.write_text(text)
            print(f"    The lammps.ini file did not exist. Created {path}")

        self.exe_config.path = path

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-lammps"

        if platform.system() == "Darwin" and platform.machine() == "arm64":
            os.environ["CONDA_SUBDIR"] = "osx-64"
        elif platform.system() == "Linux" and platform.machine() == "arm64":
            os.environ["CONDA_SUBDIR"] = "linux-64"

        super().install()

    def show(self):
        """Show the status of the LAMMPS installation."""
        print("Showing the LAMMPS installation.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "lammps.ini"
        if not path.exists():
            text = (self.resource_path / "lammps.ini").read_text()
            path.write_text(text)
            print(f"    The lammps.ini file does not exist at {path}")
            print("    The 'check' command will create it if LAMMPS is installed.")
            print("    Otherwise 'install' will install LAMMPS.")
            return

        self.exe_config.path = path

        if not self.exe_config.section_exists("local"):
            print(
                "    LAMMPS is not configured: there is no 'local' section in "
                f"     {path}."
            )
            return

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-lammps"

        super().show()

    def uninstall(self):
        """Uninstall the LAMMPS installation."""
        print("Uninstall the LAMMPS installation.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "lammps.ini"
        if not path.exists():
            text = (self.resource_path / "lammps.ini").read_text()
            path.write_text(text)
            print(
                f""""    The lammps.ini file does not exist at {path}
    Perhaps LAMMPS is not installed, but if it is the 'check' command may locate it
    and create the ini file, after which 'uninstall' will remove it."""
            )
            return

        self.exe_config.path = path

        if not self.exe_config.section_exists("local"):
            print(
                f""""    The lammps.ini file at {path} does not have local section.
    Perhaps LAMMPS is not installed, but if it is the 'check' command may locate it
    and update the ini file, after which 'uninstall' will remove it."""
            )
            return

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-lammps"

        super().uninstall()

    def update(self):
        """Updates the LAMMPS installation."""
        print("Updating the LAMMPS installation.")

        # What Conda environment is the default?
        path = self.configuration.path.parent / "lammps.ini"
        if not path.exists():
            text = (self.resource_path / "lammps.ini").read_text()
            path.write_text(text)
            print(f"    The lammps.ini file did not exist. Created {path}")

        self.exe_config.path = path

        # Get the current values
        data = self.exe_config.get_values("local")

        if "conda-environment" in data and data["conda-environment"] != "":
            self.environment = data["conda-environment"]
        else:
            self.environment = "seamm-lammps"

        super().update()

    def exe_version(self, config):
        """Get the version of the LAMMPS executable.

        Parameters
        ----------
        config : dict
            Dictionary of options for running LAMMPS

        Returns
        -------
        "LAMMPS", str
            The version reported by LAMMPS, or 'unknown'.
        """
        environment = config["conda-environment"]
        conda = config["conda"]
        if environment[0] == "~":
            environment = str(Path(environment).expanduser())
            command = f"'{conda}' run --live-stream -p '{environment}'"
        elif Path(environment).is_absolute():
            command = f"'{conda}' run --live-stream -p '{environment}'"
        else:
            command = f"'{conda}' run --live-stream -n '{environment}'"
        command += " lmp -h"
        try:
            result = subprocess.run(
                command,
                stdin=subprocess.DEVNULL,
                capture_output=True,
                text=True,
                shell=True,
            )
        except Exception:
            version = "unknown"
        else:
            version = "unknown"
            lines = result.stdout.splitlines()
            if len(lines) > 1:
                line = lines[1]
                version = " ".join(line.split()[6:])

        return "LAMMPS", version



# -*- coding: utf-8 -*-

"""Definition of unit systems in LAMMPS."""
from seamm_util import ureg, Q_, units_class  # noqa: F401

lammps_unit_system = "real"

unit_systems = {
    "real": {
        "mass": "grams/mole",
        "distance": "angstroms",
        "time": "femtoseconds",
        "energy": "kcal/mole",
        "velocity": "angstroms/femtosecond",
        "force": "kcal/mole/angstrom",
        "torque": "kcal/mole",
        "temperature": "kelvin",
        "pressure": "atmospheres",
        "dynamic viscosity": "poise",
        "charge": "elementary_charge",
        "dipole": "elementary_charge*angstroms",
        "electric field": "volts/angstrom",
        "density": "gram/cm^3",
    },
    "metal": {
        "mass": "grams/mole",
        "distance": "angstroms",
        "time": "picoseconds",
        "energy": "eV",
        "velocity": "angstroms/picosecond",
        "force": "eV/angstrom",
        "torque": "eV",
        "temperature": "kelvin",
        "pressure": "bars",
        "dynamic viscosity": "poise",
        "charge": "elementary_charge",
        "dipole": "elementary_charge*angstroms",
        "electric field": "volts/angstrom",
        "density": "gram/cm^3",
    },
    "si": {
        "mass": "kilograms",
        "distance": "meters",
        "time": "seconds",
        "energy": "joules",
        "velocity": "meters/second",
        "force": "newtons",
        "torque": "newton*meters",
        "temperature": "kelvin",
        "pressure": "pascals",
        "dynamic viscosity": "pascal*second",
        "charge": "coulombs",
        "dipole": "coulombs*meters",
        "electric field": "volts/meter",
        "density": "kilograms/meter^3",
    },
    "cgs": {
        "mass": "grams",
        "distance": "centimeters",
        "time": "seconds",
        "energy": "ergs",
        "velocity": "centimeters/second",
        "force": "dynes",
        "torque": "dyne*centimeters",
        "temperature": "kelvin",
        "pressure": "dyne/cm^2",
        "dynamic viscosity": "poise",
        "charge": "statcoulombs",
        "dipole": "statcoulombs*cm",
        "electric field": "statvolt/cm",
        "density": "grams/cm^3",
    },
    "electron": {
        "mass": "amu",
        "distance": "bohr",
        "time": "femtoseconds",
        "energy": "hartrees",
        "velocity": "bohr/a_u_time",
        "force": "hartrees/bohr",
        "temperature": "kelvin",
        "pressure": "pascals",
        "charge": "elementary_charge",
        "dipole moment": "debye",
        "electric field": "volts/cm",
    },
    "micro": {
        "mass": "picograms",
        "distance": "micrometers",
        "time": "microseconds",
        "energy": "picogram*micrometer^2/microsecond^2",
        "velocity": "micrometers/microsecond",
        "force": "picogram*micrometer/microsecond^2",
        "torque": "picogram*micrometer^2/microsecond^2",
        "temperature": "kelvin",
        "pressure": "picogram/(micrometer*microsecond^2)",
        "dynamic viscosity": "picogram/(micrometer*microsecond)",
        "charge": "picocoulombs",
        "dipole": "picocoulomb*micrometer",
        "electric field": "volt/micrometer",
        "density": "picograms/micrometer^3",
    },
    "nano": {
        "mass": "attograms",
        "distance": "nanometers",
        "time": "nanoseconds",
        "energy": "attogram*nanometer^2/nanosecond^2",
        "velocity": "nanometers/nanosecond",
        "force": "attogram*nanometer/nanosecond^2",
        "torque": "attogram*nanometer^2/nanosecond^2",
        "temperature": "kelvin",
        "pressure": "attogram/(nanometer*nanosecond^2)",
        "dynamic viscosity": "attogram/(nanometer*nanosecond)",
        "charge": "elementary_charge",
        "dipole": "elementary_charge*nanometer",
        "electric field": "volt/nanometer",
        "density": "attograms/nanometer^3",
    },
}

quantity_to_dimensionality = {
    "mass": ["[mass] / [substance]", "[mass]"],
    "distance": ["[length]"],
    "time": ["[time]"],
    "energy": [
        "[length] ** 2 * [mass] / [substance] / [time] ** 2",
        "[length] ** 2 * [mass] / [time] ** 2",
    ],
    "velocity": ["[length] / [time]"],
    "force": [
        "[length] * [mass] / [substance] / [time] ** 2",
        "[length] * [mass] / [time] ** 2",
    ],
    "torque": [
        "[length] ** 2 * [mass] / [substance] / [time] ** 2",
        "[length] ** 2 * [mass] / [time] ** 2",
    ],
    "temperature": ["[temperature]"],
    "pressure": ["[mass] / [length] / [time] ** 2"],
    "dynamic viscosity": ["[mass] / [length] / [time]"],
    "charge": ["[current] * [time]", "[length] ** 1.5 * [mass] ** 0.5 / [time]"],
    "dipole": [
        "[current] * [length] * [time]",
        "[length] ** 2.5 * [mass] ** 0.5 / [time]",
    ],
    "electric field": [
        "[length] * [mass] / [current] / [time] ** 3",
        "[mass] ** 0.5 / [length] ** 0.5 / [time]",
    ],
    "density": ["[mass] / [length] ** 3"],
    "dipole moment": ["[current] * [length] * [time]"],
}

dimensionality_to_quantity = {
    "[mass] / [substance]": ["mass"],
    "[mass]": ["mass"],
    "[length]": ["distance"],
    "[time]": ["time"],
    "[length] ** 2 * [mass] / [substance] / [time] ** 2": ["energy", "torque"],
    "[mass] * [length] ** 2 / [time] ** 2 / [substance]": ["energy", "torque"],
    "[length] ** 2 * [mass] / [time] ** 2": ["energy", "torque"],
    "[length] / [time]": ["velocity"],
    "[length] * [mass] / [substance] / [time] ** 2": ["force"],
    "[length] * [mass] / [time] ** 2": ["force"],
    "[temperature]": ["temperature"],
    "[mass] / [length] / [time] ** 2": ["pressure"],
    "[mass] / [length] / [time]": ["dynamic viscosity"],
    "[current] * [time]": ["charge"],
    "[length] ** 1.5 * [mass] ** 0.5 / [time]": ["charge"],
    "[current] * [length] * [time]": ["dipole", "dipole moment"],
    "[length] ** 2.5 * [mass] ** 0.5 / [time]": ["dipole"],
    "[length] * [mass] / [current] / [time] ** 3": ["electric field"],
    "[mass] ** 0.5 / [length] ** 0.5 / [time]": ["electric field"],
    "[mass] / [length] ** 3": ["density"],
}


def set_lammps_unit_system(value):
    """Set the default unit system for LAMMPS."""
    global lammps_unit_system
    if value not in unit_systems:
        raise ValueError(f"unit system '{value}' not recognized!")
    lammps_unit_system = value


def get_lammps_unit_system():
    """Get the default unit system for LAMMPS."""
    global lammps_unit_system
    return lammps_unit_system


def to_lammps_units(value, units=None, quantity=None, unit_system=None):
    """Convert the value with units to the current LAMMPS units."""
    if not isinstance(value, units_class):
        if units is not None:
            value = Q_(value, units)
        else:
            value = Q_(value)

    if quantity is None:
        quantities = dimensionality_to_quantity[str(value.dimensionality)]
        quantity = quantities[0]

    if unit_system is None:
        unit_system = lammps_unit_system

    unit_expr = unit_systems[unit_system][quantity]
    converted = value.to(unit_expr)

    return converted.magnitude


def from_lammps_units(value, units, quantity=None, unit_system=None):
    """Convert the scalar value from LAMMPS units to requested units"""
    if quantity is None:
        dimensions = Q_(1.0, units).dimensionality
        quantities = dimensionality_to_quantity[str(dimensions)]
        quantity = quantities[0]

    if unit_system is None:
        unit_system = lammps_unit_system

    unit_expr = unit_systems[unit_system][quantity]
    converted = Q_(value, unit_expr).to(units)

    return converted


if __name__ == "__main__":  # pragma: no cover
    import pint
    import json

    for unit in ureg:
        try:
            q = ureg(unit)
        except pint.errors.UndefinedUnitError:
            q = "unknown"
        print(f"{unit} --> {q}")
    print("")

    dimensionality = {}
    for system, units in unit_systems.items():
        print(f"{system} system of units")
        for quantity, unit_expr in units.items():
            print(f"\t{quantity} uses {unit_expr}")
            q = ureg(unit_expr)
            print(f"\t\t{q}")
            dimensions = str(q.dimensionality)
            if quantity not in dimensionality:
                dimensionality[quantity] = [dimensions]
            elif dimensions not in dimensionality[quantity]:
                dimensionality[quantity].append(dimensions)
                print(f"{system}: {quantity} <-- {unit_expr}")

    print()
    print(json.dumps(dimensionality, indent=4))

    inverse = {}
    for quantity, unit_exprs in dimensionality.items():
        for unit_expr in unit_exprs:
            if unit_expr not in inverse:
                inverse[unit_expr] = [quantity]
            elif quantity != inverse[unit_expr]:
                inverse[unit_expr].append(quantity)
                print(
                    f"quantity {quantity} != {inverse[unit_expr]} for {unit_expr}"
                )  # noqa: E501

    print()
    print(json.dumps(inverse, indent=4))

    print()
    value = "2.0 kJ/mol"
    converted = to_lammps_units(value)
    print(f"{value} in real units is {converted:.4}")
    print(f'{converted:.4} --> {from_lammps_units(converted, "kJ/mol"):~.4}')

    lammps_unit_system = "metal"
    converted = to_lammps_units(value)
    print(f"{value} in metal units is {converted:.4}")
    print(f'{converted:.4} --> {from_lammps_units(converted, "kJ/mol"):~.4}')



# -*- coding: utf-8 -*-

"""Non-graphical part of the Heat Flux step in a LAMMPS flowchart
"""

import json

import logging
from pathlib import Path
import pkg_resources

import lammps_step
from .nve import NVE
import molsystem
import seamm
from seamm_util import Q_, units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

# In addition to the normal logger, two logger-like printing facilities are
# defined: "job" and "printer". "job" send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# "printer" sends output to the file "step.out" in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("LAMMPS")

# Add this module's properties to the standard properties
path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "properties.csv"
if path.exists():
    molsystem.add_properties_from_file(csv_file)

script = """\
#!/usr/bin/env python

import argparse
import sys

from lammps import lammps, LMP_STYLE_ATOM, LMP_TYPE_VECTOR, LMP_TYPE_ARRAY
from mpi4py import MPI
import numpy as np

def J_filter_cb(lmp, ntimestep, nlocal, ids, xyz, fext, args = []):
    # Set forces to 0.0
    fext.fill(0.0)

    # lmp.fix_external_set_energy_global("J_filter", 0.0)
    # lmp.fix_external_set_virial_global("J_filter", [0.0] * 6)

    # update energy
    PE0 = lmp.numpy.extract_fix("PE_ave", LMP_STYLE_ATOM, LMP_TYPE_VECTOR)
    lmp.numpy.fix_external_set_energy_peratom("J_filter", -PE0);

    # and stress
    S_p0 = lmp.numpy.extract_fix("S_p_ave", LMP_STYLE_ATOM, LMP_TYPE_ARRAY)
    S_b0 = lmp.numpy.extract_fix("S_b_ave", LMP_STYLE_ATOM, LMP_TYPE_ARRAY)

    S = np.reshape(S_p0, (nlocal, 6))
    tmp = np.reshape(S_b0, (nlocal, 9))

    S[:, 0] += tmp[:, 0]
    S[:, 1] += tmp[:, 1]
    S[:, 2] += tmp[:, 2]
    S[:, 3] += (tmp[:, 3] + tmp[:, 6]) / 2
    S[:, 4] += (tmp[:, 4] + tmp[:, 7]) / 2
    S[:, 5] += (tmp[:, 5] + tmp[:, 8]) / 2

    lmp.numpy.fix_external_set_virial_peratom("J_filter", -S)

def run_thermal_conductivity(cmd_args=None):
    if cmd_args is None:
        lmp = lammps()
    else:
        lmp = lammps(cmdargs=cmd_args.split())
    lmp.file("input.dat")

    lmp.set_fix_external_callback("J_filter", J_filter_cb, lmp)

    lmp.file("input_post.dat")

    lmp.close()
    lmp.finalize()

# Get any arguments
parser = argparse.ArgumentParser()
parser.add_argument("--cmd-args", help="Command line arguments for LAMMPS")
kwords = vars(parser.parse_args())

run_thermal_conductivity(**kwords)
"""

script2 = """\
#!/usr/bin/env python

import argparse
import sys

from lammps import lammps, LMP_STYLE_ATOM, LMP_TYPE_VECTOR, LMP_TYPE_ARRAY
from mpi4py import MPI
import numpy as np

def J_filter_cb(lmp, ntimestep, nlocal, ids, xyz, fext, args = []):
    # Set forces to 0.0
    fext.fill(0.0)

    # update energy
    PE0 = lmp.numpy.extract_fix("PE_ave", LMP_STYLE_ATOM, LMP_TYPE_VECTOR)
    lmp.numpy.fix_external_set_energy_peratom("J_filter", -PE0);

    # and stress
    S_p0 = lmp.numpy.extract_fix("S_p_ave", LMP_STYLE_ATOM, LMP_TYPE_ARRAY)

    S = np.reshape(S_p0, (nlocal, 6))

    lmp.numpy.fix_external_set_virial_peratom("J_filter", -S)

def run_thermal_conductivity(cmd_args=None):
    if cmd_args is None:
        lmp = lammps()
    else:
        lmp = lammps(cmdargs=cmd_args.split())
    lmp.file("input.dat")

    lmp.set_fix_external_callback("J_filter", J_filter_cb, lmp)

    lmp.file("input_post.dat")

    lmp.close()
    lmp.finalize()

# Get any arguments
parser = argparse.ArgumentParser()
parser.add_argument("--cmd-args", help="Command line arguments for LAMMPS")
kwords = vars(parser.parse_args())

run_thermal_conductivity(**kwords)
"""


class HeatFlux(NVE):
    """
    The non-graphical part of a Heat Flux step in a flowchart.

    Attributes
    ----------
    parser : configargparse.ArgParser
        The parser object.

    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : HeatFluxParameters
        The control parameters for Heat Flux.

    See Also
    --------
    TkHeatFlux,
    HeatFlux, HeatFluxParameters
    """

    def __init__(
        self, flowchart=None, title="Heat Flux", extension=None, logger=logger
    ):
        """A substep for Heat Flux in a subflowchart for LAMMPS.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
        flowchart: seamm.Flowchart
            The non-graphical flowchart that contains this step.

        title: str
            The name displayed in the flowchart.
        extension: None
            Not yet implemented
        logger : Logger = logger
            The logger to use and pass to parent classes

        Returns
        -------
        None
        """
        logger.debug(f"Creating Heat Flux {self}")

        super().__init__(
            flowchart=flowchart,
            title="Heat Flux",
            extension=extension,
            logger=logger,
        )

        self._calculation = "Heat Flux"
        self._model = None
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.HeatFluxParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return lammps_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return lammps_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
        P: dict
            An optional dictionary of the current values of the control
            parameters.
        Returns
        -------
        str
            A description of the current step.
        """
        if not P:
            P = self.parameters.values_to_dict()

        text = (
            "Calculate the heat flux using a microcanonical ensemble (NVE), "
            "running for {time} with a timestep of {timestep}. The "
            "heat flux will be sampled every {heat flux}."
        )

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self, extras=None):
        """Get the input for a heat flux run in LAMMPS"""
        # See what type of forcefield we have and handle it
        ff = self.get_variable("_forcefield")
        if ff == "OpenKIM":
            ffname = ""
        else:
            ffname = ff.current_forcefield

        self.description = []

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        time = lammps_step.to_lammps_units(P["time"], quantity="time")
        timestep, P["timestep"] = self.timestep(P["timestep"])
        nsteps = round(time / timestep)

        # Work out the sampling information
        sampling = lammps_step.to_lammps_units(P["sampling"], quantity="time")
        nevery = max(1, round(sampling / timestep))
        nrepeat = 1
        nfreq = nevery * nrepeat

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description.append(
            __(self.description_text(PP), **PP, indent=3 * " ").__str__()
        )

        thermo_properties = (
            "time temp press etotal ke pe ebond "
            "eangle edihed eimp evdwl etail ecoul elong"
        )
        properties = "v_time v_temp v_press v_etotal v_ke v_pe v_emol v_epair"
        title2 = "tstep t T P Etot Eke Epe Emol Epair"

        lines = []
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append("reset_timestep      0")
        lines.append("timestep            {}".format(timestep))

        nfixes = 0
        ncomputes = 0
        ndumps = 0

        computes = []
        fixes = []

        # Unit conversion factor
        if lammps_step.get_lammps_unit_system() == "metal":
            factor = Q_("eV/Å^2/ps")
        else:
            factor = Q_("kcal/Å^2/fs/mol") / Q_("kcal/mol") * Q_("kcal/mol").to("kJ")
        factor = factor.m_as("W/m^2")

        if "cff" in ffname:
            # Centroid/stress/atom does not handle class2 ff ... cross-terms?
            lines.append(
                f"""
#          Green-Kubo method

fix                 dynamics all nve

compute             KE all ke/atom
compute             PE all pe/atom

#          centroid doesn't work with kspace, so split into pair and non-pair parts

compute             S_p all stress/atom NULL virial
compute             flux_p all heat/flux KE PE S_p

#          An initial run to settle things down.

run                 1000 post no

#          A run to get the averages for PE and S for scheme 2

fix                 ave all ave/atom 1 1000 1000 c_PE
fix                 p_ave all ave/atom 1 1000 1000 c_S_p[1] c_S_p[2] c_S_p[3] &
                                                   c_S_p[4] c_S_p[5] c_S_p[6]

run                 1000 post no

fix                 PE_ave all store/state 0 f_ave
fix                 S_p_ave all store/state 0 f_p_ave[1] f_p_ave[2] f_p_ave[3] &
                                              f_p_ave[4] f_p_ave[5] f_p_ave[6]

#          Store the initial values for scheme 1

fix                 PE0 all store/state 0 c_PE
fix                 S_p0 all store/state 0 c_S_p[1] c_S_p[2] c_S_p[3] &
                                           c_S_p[4] c_S_p[5] c_S_p[6]

#          The run is need to capture the store/state fixes

run                 0

unfix               p_ave
unfix               ave

reset_timestep      0

#          Conversion from kcal/Å^2/fs/mol to W/m^2")

variable            factor equal {factor}
variable            Jx equal v_factor*c_flux_p[1]/vol
variable            Jy equal v_factor*c_flux_p[2]/vol
variable            Jz equal v_factor*c_flux_p[3]/vol

#          The thermo output

thermo              {nsteps // 100}
thermo_style        custom {thermo_properties}

#          The external fix to adjust the peratom PE and S

fix                 J_filter all external pf/callback 1 1
fix_modify          J_filter energy yes
fix_modify          J_filter virial yes
"""
            )
        else:
            lines.append(
                f"""
#          Green-Kubo method

fix                 dynamics all nve

compute             KE all ke/atom
compute             PE all pe/atom

#          centroid doesn't work with kspace, so split into pair and non-pair parts

compute             S_p all stress/atom NULL pair kspace
compute             S_b all centroid/stress/atom NULL bond angle dihedral improper
compute             flux_p all heat/flux KE PE S_p
compute             flux_b all heat/flux KE PE S_b

#          An initial run to settle things down.

run                 1000 post no

#          A run to get the averages for PE and S for scheme 2

fix                 ave all ave/atom 1 1000 1000 c_PE
fix                 p_ave all ave/atom 1 1000 1000 c_S_p[1] c_S_p[2] c_S_p[3] &
                                                   c_S_p[4] c_S_p[5] c_S_p[6]
fix                 b_ave all ave/atom 1 1000 1000 c_S_b[1] c_S_b[2] c_S_b[3] &
                                                   c_S_b[4] c_S_b[5] c_S_b[6] &
                                                   c_S_b[7] c_S_b[8] c_S_b[9]

run                 5000 post no

dump                dPE all custom 6000 dPE.dump.* id f_ave
dump_modify         dPE sort id

variable            Sxx atom f_p_ave[1]+f_b_ave[1]
variable            Syy atom f_p_ave[2]+f_b_ave[2]
variable            Szz atom f_p_ave[3]+f_b_ave[3]
variable            Sxy atom f_p_ave[4]+(f_b_ave[4]+f_b_ave[7])/2
variable            Sxz atom f_p_ave[5]+(f_b_ave[5]+f_b_ave[8])/2
variable            Syz atom f_p_ave[6]+(f_b_ave[6]+f_b_ave[9])/2

dump                dS all custom 6000 dS.dump.* id v_Sxx v_Syy v_Szz v_Sxy v_Sxz v_Syz
dump_modify         dS sort id format float %9.0f

fix                 PE_ave all store/state 0 f_ave
fix                 S_p_ave all store/state 0 f_p_ave[1] f_p_ave[2] f_p_ave[3] &
                                              f_p_ave[4] f_p_ave[5] f_p_ave[6]
fix                 S_b_ave all store/state 0 f_b_ave[1] f_b_ave[2] f_b_ave[3] &
                                              f_b_ave[4] f_b_ave[5] f_b_ave[6] &
                                              f_b_ave[7] f_b_ave[8] f_b_ave[9]

#          Store the initial values for scheme 1

fix                 PE0 all store/state 0 c_PE
fix                 S_p0 all store/state 0 c_S_p[1] c_S_p[2] c_S_p[3] &
                                           c_S_p[4] c_S_p[5] c_S_p[6]
fix                 S_b0 all store/state 0 c_S_b[1] c_S_b[2] c_S_b[3] &
                                           c_S_b[4] c_S_b[5] c_S_b[6] &
                                           c_S_b[7] c_S_b[8] c_S_b[9]

#          The run is need to capture the store/state fixes

run                 0

undump              dPE
undump              dS

variable            Sxx delete
variable            Syy delete
variable            Szz delete
variable            Sxy delete
variable            Sxz delete
variable            Syz delete

unfix               b_ave
unfix               p_ave
unfix               ave

reset_timestep      0

#          Conversion from kcal/Å^2/fs/mol to W/m^2")

variable            factor equal {factor}
variable            Jx equal v_factor*(c_flux_p[1]+c_flux_b[1])/vol
variable            Jy equal v_factor*(c_flux_p[2]+c_flux_b[2])/vol
variable            Jz equal v_factor*(c_flux_p[3]+c_flux_b[3])/vol

#          The thermo output

thermo              {nsteps // 100}
thermo_style        custom {thermo_properties}

#          The external fix to adjust the peratom PE and S

fix                 J_filter all external pf/callback 1 1
fix_modify          J_filter energy yes
fix_modify          J_filter virial yes
"""
            )

        # instantaneous output written for averaging
        if P["sampling"] == "none":
            self.description.append(
                __(
                    "The run will be {nsteps:n} steps of dynamics.",
                    nsteps=nsteps,
                    indent=7 * " ",
                )
            )
        else:
            sampling = lammps_step.to_lammps_units(P["sampling"], quantity="time")
            nevery = max(1, round(sampling / timestep))
            nfreq = int(nsteps / nevery)
            nrepeat = 1
            nfreq = nevery * nrepeat
            nfixes += 1
            dt = (nevery * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "NVE",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // nevery,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI trajectory 2.0 " + text
            filename = f"@{self._id[-1]}+heat_flux_state.trj"
            lines.append(
                f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
                f"                       {properties} &\n"
                "                       off 2 &\n"
                f"                       title1 '{title1}' &\n"
                f"                       title2 '{title2}' &\n"
                f"                       file {filename}"
            )
            self.description.append(
                __(
                    (
                        "The run will be {nsteps:,d} steps of dynamics "
                        "sampled every {nevery:n} steps."
                    ),
                    nsteps=nsteps,
                    nevery=nevery,
                    indent=7 * " ",
                )
            )

        nevery = 10
        nfreq = int(nsteps / 10)
        nrepeat = int(nfreq / nevery)
        nfreq = nevery * nrepeat

        filename = f"@{self._id[-1]}+heat_flux_summary.trj"
        lines.append(
            f"fix                 summary all ave/time {nevery} 1 {nfreq} &\n"
            f"                       {properties} &\n"
            "                       off 2 &\n"
            f"                       file {filename}"
        )

        # Handle trajectories
        tmp, ncomputes, ndumps, nfixes = self.trajectory_input(
            P, timestep, nsteps, ncomputes, ndumps, nfixes
        )
        lines.extend(tmp)

        if extras is not None and "shake" in extras:
            lines.append(extras["shake"].format("constraint"))
            fixes.append("constraint")
        lines.append("")

        post_lines = [
            f"""

run                 {nsteps}

# Reset almost everything
thermo              0
thermo_style        one

variable            factor delete
variable            Jx delete
variable            Jy delete
variable            Jz delete
unfix               summary
unfix               J_filter
unfix               S_p0
unfix               PE0
unfix               S_p_ave
unfix               PE_ave
unfix               dynamics
"""
        ]
        for compute in computes:
            post_lines.append(f"uncompute           {compute}")
        for fix in fixes:
            post_lines.append(f"unfix               {fix}")
        for n in range(1, nfixes + 1):
            post_lines.append(f"unfix               {n}")
        for n in range(1, ncomputes + 1):
            post_lines.append(f"uncompute           {n}")
        for n in range(1, ndumps + 1):
            post_lines.append(f"undump              {n}")
        if P["heat flux"] != "never":
            if "cff" not in ffname:
                post_lines.append("uncompute           flux_b")
                post_lines.append("uncompute           S_b")
                post_lines.append("unfix               S_b0")
                post_lines.append("unfix               S_b_ave")
            post_lines.append("uncompute           flux_p")
            post_lines.append("uncompute           S_p")
            post_lines.append("uncompute           PE")
            post_lines.append("uncompute           KE")

        if "cff" in ffname:
            return {
                "script": lines,
                "postscript": post_lines,
                "use python": True,
                "python script": script2,
            }
        else:
            return {
                "script": lines,
                "postscript": post_lines,
                "use python": True,
                "python script": script,
            }



# -*- coding: utf-8 -*-
"""Control parameters for NVT (canonical) dynamics"""

import logging
import seamm

logger = logging.getLogger(__name__)


class VelocitiesParameters(seamm.Parameters):
    """The control parameters for NVT dynamics in LAMMPS"""

    parameters = {
        "method": {
            "default": "using a random distribution",
            "kind": "enumeration",
            "format_string": "s",
            "enumeration": (
                "using a random distribution",
                "scaling current velocities",
            ),
            "description": "Set the temperature",
            "help_text": (
                "The method to use to adjuust the velocities "
                "and hence the temperature."
            ),
        },
        "T": {
            "default": 298.15,
            "kind": "float",
            "default_units": "K",
            "format_string": ".2f",
            "description": "Temperature:",
            "help_text": "The temperature corresponding to the velocities.",
        },
        "seed": {
            "default": "random",
            "kind": "integer",
            "default_units": None,
            "format_string": "",
            "enumeration": ("random",),
            "description": "Random seed:",
            "help_text": (
                "The seed for the random number generator."
                "'random' means to generate a random integer "
                "as the seed."
            ),
        },
        "remove_momentum": {
            "default": (
                "remove any translational and, for molecular systems, "
                "rotational momentum (default)."
            ),
            "kind": "enumeration",
            "format_string": "s",
            "enumeration": (
                (
                    "remove any translational and, for molecular systems, "
                    "rotational momentum (default)"
                ),
                "remove translational but not rotational momentum",
                "remove rotational but not translational momentum",
                "remove both translational and rotational momentum",
                "remove neither translational nor rotational momentum",
            ),
            "description": "Momentum:",
            "help_text": (
                "Ensure that there is no overall linear "
                "momentum so the system does not translate."
            ),
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={**VelocitiesParameters.parameters, **defaults}, data=data
        )



# -*- coding: utf-8 -*-

"""NVE (microcanonical) dynamics in LAMMPS"""

import json

import lammps_step
import logging
import seamm
from seamm_util import ureg, Q_, units_class  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("lammps")


class NVE(lammps_step.Energy):
    def __init__(
        self,
        flowchart=None,
        title="NVE dynamics",
        extension=None,
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating NVE {}".format(self))

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            logger=logger,
        )
        self.logger.debug("NVE.init() creating NVE_Parameters object")

        self._calculation = "nve"
        self._model = None
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.NVE_Parameters()

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """

        if not P:
            P = self.parameters.values_to_dict()

        text = (
            "{time} of microcanonical (NVE) dynamics using a "
            "timestep of {timestep}. The trajectory will be "
            "sampled every {sampling}."
        )

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self, extras=None):
        """Get the input for an NVE dynamics run in LAMMPS"""

        # See what type of forcefield we have and handle it
        ff = self.get_variable("_forcefield")
        if ff == "OpenKIM":
            ffname = ""
        else:
            ffname = ff.current_forcefield

        self.description = []
        self.description.append(__(self.header, indent=3 * " "))

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        timestep, P["timestep"] = self.timestep(P["timestep"])

        if extras is not None and "nsteps" in extras:
            nsteps = extras["nsteps"]
        else:
            time = P["time"].to("fs").magnitude
            nsteps = round(time / timestep)

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description.append(__(self.description_text(), **PP, indent=7 * " "))

        # time = lammps_step.to_lammps_units(P['time'], quantity='time')
        # nsteps = round(time / timestep)

        thermo_properties = (
            "time temp press etotal ke pe ebond "
            "eangle edihed eimp evdwl etail ecoul elong"
        )
        properties = "v_time v_temp v_press v_etotal v_ke v_pe v_emol v_epair"
        title2 = "tstep t T P Etot Eke Epe Emol Epair"

        lines = []
        nfixes = 0
        ncomputes = 0
        ndumps = 0
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append("reset_timestep      0")
        lines.append("timestep            {}".format(timestep))
        lines.append("thermo_style        custom {}".format(thermo_properties))
        lines.append("thermo              {}".format(int(nsteps / 100)))
        nfixes += 1
        lines.append("fix                 {} all nve".format(nfixes))

        # For the heat flux, if requested, we need extra input
        if P["heat flux"] != "never":
            # Unit conversion factor
            if lammps_step.get_lammps_unit_system() == "metal":
                factor = Q_("eV/Å^2/ps")
            else:
                factor = (
                    Q_("kcal/Å^2/fs/mol") / Q_("kcal/mol") * Q_("kcal/mol").to("kJ")
                )
            factor = factor.m_as("W/m^2")
            if "cff" in ffname or not P["use centroid stress"]:
                # Centroid/stress/atom does not handle class2 ff ... cross-terms?
                lines.append(
                    f"""
compute             KE all ke/atom
compute             PE all pe/atom

compute             S_p all stress/atom NULL virial
compute             flux_p all heat/flux KE PE S_p

#          Conversion from kcal/Å^2/fs/mol to W/m^2

variable            factor equal {factor}
variable            Jx equal v_factor*c_flux_p[1]/vol
variable            Jy equal v_factor*c_flux_p[2]/vol
variable            Jz equal v_factor*c_flux_p[3]/vol
"""
                )
            else:
                lines.append(
                    f"""
compute             KE all ke/atom
compute             PE all pe/atom

#          centroid doesn't work with kspace, so split into pair and non-pair parts

compute             S_p all stress/atom NULL pair kspace
compute             S_b all centroid/stress/atom NULL bond angle dihedral improper
compute             flux_p all heat/flux KE PE S_p
compute             flux_b all heat/flux KE PE S_b

#          Conversion from kcal/Å^2/fs/mol to W/m^2

variable            factor equal {factor}
variable            Jx equal v_factor*(c_flux_p[1]+c_flux_b[1])/vol
variable            Jy equal v_factor*(c_flux_p[2]+c_flux_b[2])/vol
variable            Jz equal v_factor*(c_flux_p[3]+c_flux_b[3])/vol
"""
                )

        # summary output written 10 times during run so we can see progress
        nevery = 10
        nfreq = int(nsteps / 10)
        nrepeat = int(nfreq / nevery)
        nfreq = nevery * nrepeat
        nfixes += 1
        filename = f"@{self._id[-1]}+nve_summary.trj"
        lines.append(
            f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
            f"                       {properties} &\n"
            "                       off 2 &\n"
            f"                       file {filename}"
        )

        # instantaneous output written for averaging
        if P["sampling"] == "none":
            self.description.append(
                __(
                    "The run will be {nsteps:n} steps of dynamics.",
                    nsteps=nsteps,
                    indent=7 * " ",
                )
            )
        else:
            sampling = lammps_step.to_lammps_units(P["sampling"], quantity="time")
            nevery = max(1, round(sampling / timestep))
            nfreq = int(nsteps / nevery)
            nrepeat = 1
            nfreq = nevery * nrepeat
            nfixes += 1
            dt = (nevery * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "NVE",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // nevery,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI trajectory 2.0 " + text
            filename = f"@{self._id[-1]}+nve_state.trj"
            lines.append(
                f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
                f"                       {properties} &\n"
                "                       off 2 &\n"
                f"                       title1 '{title1}' &\n"
                f"                       title2 '{title2}' &\n"
                f"                       file {filename}"
            )
            self.description.append(
                __(
                    (
                        "The run will be {nsteps:,d} steps of dynamics "
                        "sampled every {nevery:n} steps."
                    ),
                    nsteps=nsteps,
                    nevery=nevery,
                    indent=7 * " ",
                )
            )

        # Handle trajectories
        tmp, ncomputes, ndumps, nfixes = self.trajectory_input(
            P, timestep, nsteps, ncomputes, ndumps, nfixes
        )
        lines.extend(tmp)

        if extras is not None and "shake" in extras:
            nfixes += 1
            lines.append(extras["shake"].format(nfixes))

        lines.append("")
        lines.append("run                 {}".format(nsteps))
        lines.append("")

        for i in range(1, ncomputes + 1):
            lines.append(f"uncompute           {i}")
        for i in range(1, ndumps + 1):
            lines.append(f"undump              {i}")
        for i in range(1, nfixes + 1):
            lines.append(f"unfix               {i}")
        if P["heat flux"] != "never":
            if "cff" not in ffname and P["use centroid stress"]:
                lines.append("uncompute           flux_b")
                lines.append("uncompute           S_b")
            lines.append("uncompute           flux_p")
            lines.append("uncompute           S_p")
            lines.append("uncompute           PE")
            lines.append("uncompute           KE")
            lines.append("variable            factor delete")
            lines.append("variable            Jx delete")
            lines.append("variable            Jy delete")
            lines.append("variable            Jz delete")
        lines.append("")

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }

    def timestep(self, value):
        """Get the timestep in the correct units.

        This handles the 'normal', 'accurate' and 'coarse' values,
        which depend on the mass in an empirical fashion.

        Parameters
        ----------
        value : str or Pint quantity
            The desired timestep, which may be a Pint quantity with units
            or one of 'notmal', 'accurate but slow', or 'coarse but fast'

        Returns
        -------
        timestep : float
            The magnitude of the time step in the appropriate LAMMPS units
        """
        masses = self.parent._data["masses"]
        min_mass = min(masses)

        # These are based on masses as a proxy for vibrational frequencies
        if min_mass < 10:
            factor = 1
        elif min_mass < 50:
            factor = 2
        else:
            factor = 4

        if value == "normal":
            timestep = 1.0 * factor
            value = Q_(timestep, ureg.fs)
        elif value == "accurate but slow":
            timestep = 0.5 * factor
            value = Q_(timestep, ureg.fs)
        elif value == "coarse but fast":
            timestep = 2.0 * factor
            value = Q_(timestep, ureg.fs)

        timestep = lammps_step.to_lammps_units(value, quantity="time")

        return (timestep, value)

    def trajectory_input(self, P, timestep, nsteps, ncomputes, ndumps, nfixes):
        """Create the part of the input handling the trajectories.

        Parameters
        ----------
        P : dict
            The dictionary of options
        timestep : int
            The timestep in LAMMPS units
        nsteps : int
            Total number of steps in the run
        ncomputes : int
            The counter for the computes
        ndumps : int
            The counter for the dumps
        nfixes : int
            The counter for the fixes

        Returns
        -------
        [str]
            The input lines as array
        int
            The counter for the computes
        int
            The counter for the dumps
        int
            The counter for the fixes
        """
        lines = []
        if P["atomic positions"] != "never":
            t_s = lammps_step.to_lammps_units(P["atomic positions"], quantity="time")
            n = max(1, round(t_s / timestep))
            ndumps += 1
            filename = f"@{self._id[-1]}+atomic_positions.dump_trj"
            lines.append(
                "\n"
                f"dump                {ndumps} all custom {n} {filename} id xu yu zu\n"
                f"dump_modify         {ndumps} sort id"
            )
        if P["com positions"] != "never":
            t_s = lammps_step.to_lammps_units(P["com positions"], quantity="time")
            n = max(1, round(t_s / timestep))
            filename = f"@{self._id[-1]}+com_positions.trj"
            ncomputes += 1
            c1 = ncomputes
            ncomputes += 1
            c2 = ncomputes
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "com positions",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI vector_trajectory 2.0 " + text
            title2 = "! timestep n_molecules"
            title3 = "molecule com_x com_y com_z"
            lines.append(
                "\n"
                f"compute             {c1} all chunk/atom molecule\n"
                f"compute             {c2} all com/chunk {c1}\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} c_{c2}[*] &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        title3 '{title3}' &\n"
                f"                        file {filename} mode vector"
            )
        if P["atomic velocities"] != "never":
            t_s = lammps_step.to_lammps_units(P["atomic velocities"], quantity="time")
            n = max(1, round(t_s / timestep))
            ndumps += 1
            filename = f"@{self._id[-1]}+atomic_velocities.dump_trj"
            lines.append(
                "\n"
                f"dump                {ndumps} all custom {n} {filename} id vx vy vz\n"
                f"dump_modify         {ndumps} sort id"
            )
        if P["com velocities"] != "never":
            t_s = lammps_step.to_lammps_units(P["com velocities"], quantity="time")
            n = max(1, round(t_s / timestep))
            filename = f"@{self._id[-1]}+com_velocities.trj"
            ncomputes += 1
            c1 = ncomputes
            ncomputes += 1
            c2 = ncomputes
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "com velocities",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI vector_trajectory 2.0 " + text
            title2 = "! timestep n_molecules"
            title3 = "molecule com_vx com_vy com_vz"
            lines.append(
                "\n"
                f"compute             {c1} all chunk/atom molecule\n"
                f"compute             {c2} all vcm/chunk {c1}\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} c_{c2}[*] &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        title3 '{title3}' &\n"
                f"                        file {filename} mode vector"
            )
        if P["heat flux"] != "never":
            t_s = lammps_step.to_lammps_units(P["heat flux"], quantity="time")
            n = max(1, round(t_s / timestep))
            filename = f"@{self._id[-1]}+heat_flux.trj"
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "heat flux",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI trajectory 2.0 " + text
            title2 = "Jx Jy Jz"
            lines.append(
                "\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} "
                "v_Jx v_Jy v_Jz &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        file {filename}"
            )
        if P["shear stress"] != "never":
            t_s = lammps_step.to_lammps_units(P["shear stress"], quantity="time")
            n = max(1, round(t_s / timestep))
            filename = f"@{self._id[-1]}+shear_stress.trj"
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "shear stress",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI trajectory 2.0 " + text
            title2 = "Pxy Pxz Pyz"
            lines.append(
                "\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} "
                "v_pxy v_pxz v_pyz &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        file {filename}"
            )

        return lines, ncomputes, ndumps, nfixes
